# multichannel.py - Muxer for sockets and pipes
#
# Copyright 2008 Antti Kervinen (ask@cs.tut.fi)
#
# This file is part of Ruis.
#
# Ruis is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/lgpl.html>.

"""
Socket muxer

Example:

    Client:

    When mcs is a MultiChannelSocket object, this sends HELO through a
    new channel and receives a message is an answer to the HELO
    message.

    def foo(): pass

    mcs=MultiChannelSocket(s,start_talk=1,callback=foo)
    answer=mcs.request(DirspecReqMsg('GREE'))

    Server:

    mcs=MultiChannelSocket(s,start_talk=0,callback=server)

    def server(mcs,msg,channel_id):
        if type(msg)==DirspecReqMsg:
            mcs.send(AnswerMsg('TINGS'),channel_id)


"""

import thread
import types
import socket

try: import cPickle
except:
    try:
        import pickle
        cPickle = pickle # use quietly pickle instead of cPickle
    except:
        raise ImportError("Cannot import cPickle or pickle!")

class _Channel(object):
    def __init__(self,identifier):
        self.identifier=identifier
        self.message=None
        self.readlock=thread.allocate_lock()
        # readlock is released by MultiChannelSocket._reader (and
        # no-one else) when a message arrives. Lock is acquired
        # immediately by the handler
        self.readlock.acquire()

class ConnectionLost(object):
    def __init__(self,s): self.msg=s
    def __str__(self): return str(self.msg)

class MultiChannelSocket(object):

    def __init__(self,socket_object,start_talk,default_msg_handler):
        """
        Start_talk must be 0 or 1 and different on both ends of the
        channel.

        default_msg_handler is a function that takes three parameters:
        1) multichannelsocket object mcs 2) received message 3) the
        number of the channel to which a response can be sent with
        mcs.send.
        """
        self._socket=socket_object
        self._socket_write_lock=thread.allocate_lock()
        self._msg_handlers={}
        self._msg_handlers_lock=thread.allocate_lock()
        self._all_channels={}
        self._free_channels=[]
        self._channels_lock=thread.allocate_lock()
        self._next_channel=start_talk
        self._default_msg_handler=default_msg_handler
        self._reader_thread=thread.start_new_thread(self._reader,())

    def send(self,msg,channel_identifier=None):
        """
        Sends message msg through the channel. If the channel is not
        given, an unused (possibly a new) channel is used. Function
        returns the channel through which the message was sent.
        """
        if channel_identifier==None:
            self._channels_lock.acquire()
            if len(self._free_channels)==0: # no free channels => create new
                channel_identifier=self._next_channel
                channel=_Channel(channel_identifier)
                self._all_channels[channel_identifier]=channel
                self._next_channel+=2
            else: # reuse a free channel
                channel=self._free_channels.pop()
                channel_identifier=channel.identifier
            self._channels_lock.release()

        msg.MCS_channel=channel_identifier
        
        try: pmsg=cPickle.dumps(msg)
        except:
            print "Failed to pickle object %s" % msg
            pmsg=cPickle.dumps("MULTICHANNEL: PICKLE FAILED")
        self._socket_write_lock.acquire()
        try: self._socket.send("%s %s" % (len(pmsg),pmsg))
        except socket.error, e: channel_identifier=None
        self._socket_write_lock.release()
        
        return channel_identifier

    def recv(self,channel_identifier):
        """
        Receives a message that comes from the given channel. If
        channel is not specified, receives message whose channel is
        not listened. Returns pair (message, channel).

        NOTE!!! If you use this directly, the returned channel is not
        freed. Use this only for server.
        """
        if not channel_identifier in self._all_channels or \
            channel_identifier in self._free_channels:
            raise Exception("recv called for a non-used channel!")
        else:
            channel=self._all_channels[channel_identifier]
        channel.readlock.acquire() # wait for _reader to release the lock
        return channel.message, channel.identifier

    def read_answer(self,channel_identifier):
        if channel_identifier==None: return ConnectionLost("sending failed.")
        try: response,channel_identifier=self.recv(channel_identifier)
        except socket.error, e: response=ConnectionLost(e)
        self._channels_lock.acquire()
        self._free_channels.append(self._all_channels[channel_identifier])
        self._channels_lock.release()
        return response

    def request(self,msg):
        """
        Sends a message and returns the received answer. The used
        channel is automatically allocated and freed. Use this for
        client communication.
        """
        response=self.read_answer(self.send(msg))
        return response

    def async_request(self,msg):
        """
        Sends a message but does not wait for an answer.
        """
        # FIX: do this without new thread: just send
        # a message without reserving yet another channel
        channel_id=self.send(msg)
        if channel_id!=None:
            thread.start_new_thread(self.read_answer,(channel_id,))

    def _reader(self):
        """
        This is for internal use only. It is the only method that
        reads the socket. This method runs in its own thread.
        """
        s=self._socket
        readbuffer=""
        while True:
            try: newstr=s.recv(4096)
            except: newstr=""
            if len(newstr)==0:
                # Connection lost. Tell it to the default msg handler
                self._default_msg_handler(None,None,None)
                # release all read locks to prevent deadlocking.
                # reading returns now disconnected-objects
                for c in self._all_channels:
                    if c not in self._free_channels:
                        chn=self._all_channels[c]
                        chn.message=ConnectionLost("Return value missing.")
                        chn.readlock.release()
                break # this thread quits
            readbuffer+=newstr
            while len(readbuffer)>0: # build all received messages
                msg_fields=readbuffer.split(" ",1)
                if len(msg_fields)<2: break # readbuffer does not contain a msg
                bytes_in_msg=int(msg_fields[0])
                header_length=len(msg_fields[0])+1
                if len(readbuffer)-header_length>=bytes_in_msg:
                    # extract one message from buffer
                    smsg=readbuffer[header_length:header_length+bytes_in_msg]
                    readbuffer=readbuffer[header_length+bytes_in_msg:]
                    msg=cPickle.loads(smsg)
                    if not hasattr(msg,'MCS_channel'):
                        raise Exception("multichannel received garbage: %s, orig: '%s'" %
                                        (str(msg), smsg))
                    if msg.MCS_channel in self._all_channels and \
                       msg.MCS_channel not in self._free_channels:
                        # someone is waiting for this message
                        chn=self._all_channels[msg.MCS_channel]
                        chn.message=msg
                        chn.readlock.release()
                    else:
                        # no-one waiting, message is sent to the
                        # default handler

                        # FIX: a Homo sapience would use a thread pool
                        # here, but a Neanderthal starts a new thread
                        # every time.
                        thread.start_new_thread(self._default_msg_handler,
                                                (self,msg,msg.MCS_channel))
                else:
                    break # more data needed for this message
