# ruis.py - Distributed programming platform for Python
#
# Copyright 2008 Antti Kervinen (ask@cs.tut.fi)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/lgpl.html>.


"""
Examples:

# generate a new and empty ruis (virtual file system)
import ruis
fs=ruis.newfs()
print fs['/'].ls()

# merge directory structure in /usr/ruis/basefs to the fs
ruis.load_directory(fs, '/usr/ruis/basefs')

# generate a new fs and merge basefs and devices directories to it
fs=ruis.newfs('/usr/ruis/basefs','/usr/ruis/devices')
"""

import sys
import os
import types
import aspects
import thread
from multichannel import ConnectionLost
import re

version = "0.1.3"

_CO_FLAGS_SINGLESTARARG = 0x04
_CO_FLAGS_DOUBLESTARARG = 0x08

def debugmsg_printer(prefix):
    def debugmsg(msg):
        global debug_level
        if debug_level > 0:
            sys.stderr.write('DBG %s: %s\n' % (prefix, msg))
    return debugmsg

def warning_printer(prefix):
    def warningmsg(msg):
        sys.stderr.write('WARNING %s: %s\n' % (prefix, msg))
    return warningmsg

debugmsg = debugmsg_printer("RUIS")

debug_level = 0

# Naming:
#
# dirspec - (string) a ruis path, for example "rp/0/dev"
# fso     - a file system object (inherited from DirType, FileType)

# Automatically search imported ruis file systems from site-packages
# directories
RUISPATH = os.getenv('RUISPATH','.').split(':') + \
    [p for p in sys.path if p[-14:]=='/site-packages']

########################################################################
#
# File system functions
#
########################################################################

def newfs(*pathlist):
    """
    Creates and returns the root directory of a new ruis file
    system. Directories (strings) in pathlist are merged to the
    returned file system. The directories are searched from RUISPATH.
    If pathlist element is a pair, the first element is the directory
    and the second is the modfilter that specifies the modules which
    should be loaded from the directory (see load_directory).
    """
    fs=DirType(name="")
    fs.fullname=[""]
    for loadspec in pathlist:
        if type(loadspec) in [tuple, list] and len(loadspec) == 2:
            p, modfilter = loadspec
        elif type(loadspec) == str:
            p, modfilter = loadspec, ""
        else:
            raise Exception("Invalid path list element: %s" % (loadspec,))
        load_directory(fs, p, modfilter)
    return fs

def load_directory(basefs, p, modfilter=""):
    """
    Resolves absolute name for path p, which may be located in
    RUISPATH, and uses _load_directory to load the objects in that
    path. See the documentation of _load_directory.
    """
    if not type(p) == str or not len(p)>0:
        raise Exception("Invalid directory: '%s'" % (p,))

    if not p.startswith(os.path.sep): # p is a relative path
        # try to prefix the relative path with the directories in RUISPATH
        for r in RUISPATH:
            dir_candidate = r + os.path.sep + p
            if os.path.isdir(dir_candidate):
                p = dir_candidate # now p is an (existing) absolute path
                break
        else: # for loop was not breaked out => no matching directory
            raise Exception("Cannot find directory '%s' in RUISPATH '%s'" % (p, ":".join(RUISPATH)))
    elif not os.path.isdir(p):
        raise Exception("Directory '%s' does not exist" % p)
    
    if debug_level > 0:
        debugmsg("loading directory '%s', filter='%s'" % (os.path.realpath(p), modfilter))
    _load_directory(basefs, p, modfilter)
    

def _load_directory(basefs, fsdir, modfilter="", _rootfs=None, _ruisdirname=""):
    """
    Imports files and directories in fsdir to basefs. Basefs is a Ruis
    file system. Fsdir is a path (string) to the directory whose
    contents will be loaded to the basefs recursively. Subdirectories
    are loaded first (in lexical order) and then modules in the
    current directory (in lexical order).
    
    modfilter is a regular expression that is matched to ruis module
    names with full ruis path. Only the matching modules are
    loaded. re.match is used directly, so
    modfilter="/conn/rp|.*/webformp" will match modules "/conn/rp",
    "/conn/rpxy" and "/conn/webformp". Use dollar-sign to match the end
    of string: modfilter="/conn/rp$|/conn/webformp$" will not load
    "/conn/rpxy". Names of modules in the root directory are prefixed
    by '/'.
    """
    if _rootfs == None: _rootfs = basefs
    
    if modfilter != "" and type(modfilter)==str:
        modfilter = re.compile(modfilter)

    origdir = os.getcwd()
    os.chdir(fsdir)
    
    _dirpath, _dirnames, _filenames = os.walk('.').next()
    _dirnames.sort()
    _filenames.sort()
    
    for dname in _dirnames:
        # directories starting with dot (.) are skipped
        if dname[0]==".": continue
        if not dname in basefs.ls():
            basefs.create(DirType(name=dname))
        _load_directory(basefs[dname], dname,
                       modfilter = modfilter,
                       _rootfs = _rootfs,
                       _ruisdirname = "%s/%s" % (_ruisdirname, dname))

    new_globals = {}
    sys.path.append(os.getcwd())
    for fname in _filenames:
        if fname[:1] == '_': continue
        if fname[-3:] != ".py": continue
        if modfilter != "" and \
                not modfilter.match(_ruisdirname + "/" + fname[:-3]):
            if debug_level > 0:
                debugmsg("skipping module '%s' (filtered out)" % (_ruisdirname + "/" + fname[:-3],))
            continue
        fname = fname[:-3] # remove trailing .py
        if debug_level > 0:
            debugmsg("loading module '%s'" % (_ruisdirname + "/" + fname,))
        mod_f = __import__(fname, new_globals, {}, [])
        if fname in mod_f.__dict__ and callable(mod_f.__dict__[fname]):
            basefs.create(mod_f.__dict__[fname](_rootfs))
        else:
            raise Exception("File %s does not include callable object '%s'."
                            % (mod_f, fname))
    sys.path.pop()

    os.chdir(origdir)


# Consider: perhaps async should be implemented as a fso in a /threads
# directory?
def async(fun, *args, **kwargs):
    thread.start_new_thread(fun, args, kwargs)


########################################################################
#
# FILE SYSTEM OBJECTS (FSOs)
#
########################################################################

class FSOType(object):
    def __init__(self, *args, **kwargs):
        # Set self.name attribute
        if 'name' in kwargs: self.name = kwargs['name']
        else: self.name=self.__class__.__name__.lower()
                    
    def __str__(self):
        return self.name


class FileType(FSOType):
    def __init__(self, *args, **kwargs):
        FSOType.__init__(self, *args, **kwargs)

    def __getitem__(self, dirspec):
        # split the dirspec in two: the first and the rest dirs
        splitted = _split_dirspec(dirspec)

        if splitted[0] == "": return self

        # events is a special subdirectory of every file object
        if splitted[0] == 'events':
            self._events = DirType()
            self._events.fullname = [self.fullname[0] + "/events"]
            # add methods to self._events directory
            for m in dir(self):
                if m[:1] == "_": continue # skip methods that begin with _
                if type(getattr(self, m)) == types.MethodType:
                    self._events.create(
                        EventType('call:%s' % m, getattr(self, m)),
                        "call:%s" % m)
                    self._events.create(
                        EventType('return:%s' % m, getattr(self, m)),
                        "return:%s" % m)
            return self._events[splitted[1]]
        # Here splitted[0] is not in ["", "events"]. Because file
        # object does not have other subdirectories, __getitem__
        # request is invalid.
        return None
        

### File system objects: DIRECTORIES

class DirType(FileType):
    """
    Any local directory
    """
    def __init__(self, *args, **kwargs):
        FileType.__init__(self, *args, **kwargs)
        self._contents = {} # str -> FSOType

    def create(self, fso, name = None):
        if name == None:
            regname = str(fso)
            # regname = regname[regname.rfind('.') + 1:]
        else:
            regname = name
        return self._register_to_dir(regname, fso)

    def remove(self, fso=None):
        """
        Fso is either the name of an object or the object in the
        directory. Returns the name of the removed object in the
        directory (which may be different from the name attribute of
        the object) or None if the object was not found in the
        directory.
        """
        if fso==None:
            # this item is going to be removed because its directory
            # is about to do so.
            # FIX: should this automatically remove all entries in the
            # dir and call .remove() recursively?
            # This time, no further actions taken.
            return
        # FIX: this is not a real lock...
        retval = None
        if hasattr(self, '_removelock_%s' % fso): return
        setattr(self, '_removelock_%s' % fso, 1)
        if fso in self._contents:
            del self._contents[fso]
            retval = fso
        else:
            for k, v in self._contents.iteritems():
                if v == fso:
                    fso.remove()
                    del self._contents[k]
                    retval = k
        delattr(self,'_removelock_%s' % fso)
        return retval

    def ls(self):
        l=self._contents.keys()
        l.sort()
        return tuple(l)
        
    def __getitem__(self,dirspec):
        if type(dirspec)!=str:
            if hasattr(dirspec,"fullname"): dirspec=dirspec.fullname[0]
            else: dirspec=str(dirspec)
        filereturn=FileType.__getitem__(self,dirspec)
        if filereturn!=None: return filereturn
        # remove extra slashes
        splitted=_split_dirspec(dirspec)
        if splitted[0]=="": return self

        if not splitted[0] in self._contents:
            # requested entry cannot be found in this dir!
            return None
        else:
            # search recursively from subdir splitted[0]
            return self._contents[splitted[0]][splitted[1]]
    
    def _register_to_dir(self,key,value):
        """
        Add new FSOType object (value) to this directory with given
        string label (key). Set value.fullname attribute to specify
        the list of absolute addresses in this ruisfw.
        """
        self._contents[key]=value
        if not isinstance(value, FSOType):
            raise Exception("Invalid object to be added in directory %s (%s): type=%s, value=%s" %
                            (self.name, str(self.fullname), type(value), str(value)))
        if not hasattr(value,"fullname"):
            value.fullname=[]
        value.fullname.extend( ["%s/%s" % (p,key) for p in self.fullname] )
        return key

### File system objects: EVENTS

class EventType(FSOType):
    def __init__(self, eventspec, method):
        self._eventspec = eventspec
        self._method = method

    def __getitem__(self, dirspec):
        if dirspec != '':
            raise Exception("Event object used as a directory: %s/%s"
                            % (str(self), dirspec))
        return self
    
    def set_callback(self, func):
        # debugmsg('set_callback: %s // %s // %s' % (type(func), func, dir(func)))
        if "call:" in self._eventspec:
            return aspects.wrap_around(self._method,
                                       _generate_call_advice(func),
                                       instances = [self._method.im_self])
        elif "return:" in self._eventspec:
            return aspects.wrap_around(self._method,
                                       _generate_return_advice(func),
                                       instances = [self._method.im_self])

    def enable_callback(self, callback_id):
        return aspects.enable_wrap(self._method,callback_id)

    def disable_callback(self, callback_id):
        return aspects.disable_wrap(self._method,callback_id)

    def get_sync(self):
        syncfun,adv = _generate_sync_and_advice(self._eventspec[0])
        aspects.wrap_around(self._method, adv, instances = [self._method.im_self])
        return syncfun

    def get_paramnames(self):
        code = self._method.func_code
        rv = list(code.co_varnames[:code.co_argcount])
        lastarg = code.co_argcount
        if code.co_flags & _CO_FLAGS_SINGLESTARARG:
            rv.append('*%s' % (code.co_varnames[lastarg],))
            lastarg+=1
        if code.co_flags & _CO_FLAGS_DOUBLESTARARG:
            rv.append('**%s' % (code.co_varnames[lastarg],))
        return tuple(rv)

# Helpers for Events: generate advices compatible with aspects library
# from normal functions:

def _generate_sync_and_advice(c_or_r):
    """
    generates a function that executes synchronously with call
    (parameter 'c') or return (parameter 'r') events.
    """
    lock1=thread.allocate_lock()
    lock1.acquire()
    lock2=thread.allocate_lock()
    lock2.acquire()
    messagequeue=[]
    def sync_function():
        lock1.acquire()
        lock2.release()
        return messagequeue.pop(0)
    def return_sync(retval,*a,**kw):
        messagequeue.append((retval,a,kw))
        lock1.release()
        lock2.acquire()
    def call_sync(*a,**kw):
        messagequeue.append((a,kw))
        lock1.release()
        lock2.acquire()
    if c_or_r=='c':
        return sync_function, _generate_call_advice(call_sync)
    elif c_or_r=='r':
        return sync_function, _generate_return_advice(return_sync)
    else: raise Exception("Argument error, should be 'c' or 'r', got: '%s'" % c_or_r)

def _generate_call_advice(func):
    def call_adv(self,*args,**kwargs):
        retval=func(*args,**kwargs)
        if retval==None:
            # callback returns None, call original method normally
            #print "call adv: calling as is."
            return self.__proceed(*args,**kwargs)
        elif type(retval)==tuple and len(retval)==2:
            # assume retval includes new arguments
            #print "call adv: calling with new args"
            return self.__proceed(*retval[0], **retval[1])
        else: # FIX!!! ADD THIS: elif type(retval)==multichannel.ConnectionLost
            # skip calling the real method
            #print "call adv: call skipped, retval=",retval
            return None
    return call_adv

def _generate_return_advice(func):
    def return_adv(self,*args,**kwargs):
        orig_retval=self.__proceed(*args,**kwargs)
        retval=func(orig_retval,*args,**kwargs)
        if retval==None:
            # callback returns None, return original return value
            return orig_retval
        elif type(retval)==tuple and len(retval)==2 and retval[0]=='force retval':
            # changing the return value
            return retval[1]
        else:
            return retval
    return return_adv

def _split_dirspec(dirspec):
    """
    Returns a pair of strings: a string before the first slash in
    dirspec, where the slashes starting the dirspec are not counted,
    and a string after it. If there is no slash, the seconds string is
    empty and the first contains the dirspec.
    """
    # remove starting slashes
    while len(dirspec)>0 and dirspec[0]=='/': dirspec=dirspec[1:]
    splitted=dirspec.split("/",1)
    if len(splitted)==1:
        splitted.append("")
    return splitted


########################################################################
#
# Types of messages between ruis platforms
#
########################################################################

class CarrierMsg(object):
    def __init__(self,payload):
        self.payload=payload

class ProxyMsg(object):
    """
    This message carries a proxy of the original object orig.  Methods
    of the proxy object must be created after receiving the message.
    """
    def __init__(self,orig):
        self.method_names=[ m for m in dir(orig)
                            if type(getattr(orig,m))==types.MethodType or \
                            type(getattr(orig,m))==types.FunctionType ]
        if hasattr(orig,'fullname'):
            self.remote_fullname=orig.fullname
        else:
            self.remote_id=id(orig)
    def __getitem__(self,*args):
        return self.__dict__["__getitem__"](*args)

class DirspecReqMsg(object):
    """
    This message requests a dirspec from a remote peer. For example,
    requesting fs['/conn/rp/0/dev/console'] causes
    DirspecReq('/dev/console') to be sent to the peer at /conn/rp/0.
    """
    def __init__(self,dirspec):
        self.dirspec=dirspec

class ForwardedCallMsg(object):
    def __init__(self,object_id,method_name,args,kwargs):
        self.object_id=object_id
        self.method_name=method_name
        self.args=args
        self.kwargs=kwargs

class FuncProxy(object):
    """
    A proxy of function 'orig', usually carried in ForwardedCallMsg().args
    Could this be merged to ProxyMsg?
    """
    def __init__(self,orig):
        self.func_id=orig
        self.func_channel=None # this is set where FuncProxy is reconstructed.
    def __call__(self,*args,**kwargs):
        # FIX: fix arguments again so that FuncProxies are sent
        # instead of functions (see call_forwarder in rp.py)
        fwdcall=ForwardedCallMsg(None,self.func_id,args,kwargs)
        retval=self.func_channel.request(fwdcall)
        if type(retval)==CarrierMsg:
            return retval.payload

# class ModuleTransporter(object):
#     """
#     This object stores the information needed to reconstruct a module:
#     name of the module and the source code.
#     """
#     def __init__(self,python_module):
#         self._find_source(python_module)
#         self.name=python_module.__name__
        
#     def _find_source(self,python_module):
#         srcfile=python_module.__file__
#         if srcfile.endswith("pyc"): srcfile=srcfile[:-1] # .pyc -> .py
#         f=None
#         try: f=file(srcfile)
#         except: pass
#         if not f:
#             for p in sys.path:
#                 try:
#                     if p=="": f=file(srcfile)
#                     else: f=file("%s/%s" % (p,srcfile))
#                     break
#                 except: pass
#             else:
#                 self.source=""
#             raise Exception("Cannot find file '%s'" % srcfile)
#         self.source=f.read()
#         f.close()


########################################################################
#
# Advices (wraps) for mutual exclusion
#
########################################################################

def lock_method_in_objlevel(method):
    # FIXME: Does not work / is this even needed?
    # TODO: Finish and use, if necessary
    lock_attribute_name = "_objlevel_lock_%s" % (method.__name__,)
    def allocate_lock_for_obj(self, *args, **kwargs):
        aspects.proceed
        setattr(self, lock_attribute_name, thread.allocate_lock())
    def locking_wrap(self, *args, **kwargs):
        getattr(self, lock_attribute_name).acquire()
        yield aspects.proceed
        getattr(self, lock_attribute_name).release()
    # TODO: WRAP ALLOCATE_LOCK AROUND INIT AND...
    aspects.with_wrap(locking_wrap, method, instances)
