# rp.py - Ruis protocol for connecting Ruis filesystems
#
# Copyright 2008 Antti Kervinen (ask@cs.tut.fi)
#
# This file is part of Ruis.
#
# Ruis is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/lgpl.html>.

"""
rp connection type for connecting two ruis file systems
"""

from ruis import DirType, \
     ProxyMsg, ForwardedCallMsg, DirspecReqMsg, CarrierMsg, \
     FuncProxy, debugmsg_printer

import socket
from multichannel import MultiChannelSocket, ConnectionLost, cPickle
import thread
import types
import inspect
import sys
import aspects

debugmsg = debugmsg_printer('RP')

# proxytable stores objects whose proxies have been sent
# to other ruis file systems.
_proxytable={} # map proxy id to proxy object or function

class RuisProtocolException(Exception): pass

class rp(DirType):
    """
    Implements /conn/rp
    """
    def __init__(self,local_root):
        """
        Local root is needed, because rp connections must be
        provided access to the local virtual file system.
        """
        DirType.__init__(self)
        self.name="rp"
        self.local_root=local_root
        self._autoaccept=0
        self._sockets={}

    def create(self,connspec,**kwargs):
        """
        /conn/rp.create([connspec[, local_root]])

        connspec can be
        
        - URL reference (connect to): 'tcp://localhost:5555'

        - Listen to a tcp port 'listen:tcp:5555'

        - standard input and output of a user-defined command:
          'stdio://ssh localhost 'stdio-ruis-server'' (TODO)
        
        local_root is the directory shown as this peer's root
        directory to the remote peer.
        """
        if not 'local_root' in kwargs:
            kwargs['local_root']=self.local_root

        # is this "listen.*" URI?
        if connspec[:11]=="listen:tcp:":
            try: host,port=connspec[10:].split(":")
            except e:
                raise RuisProtocolException("Error in listen URL: %s" % e)
            conn=self._accept_tcp_connection(int(port))
            rpc=RPConn(1,open_socket=conn,**kwargs)
        elif str(connspec)[:6]=="tcp://" and \
                 ":" in str(connspec)[7:]:
            # create connection based on URL tcp://host:port
            host,port = str(connspec)[6:].split(":")
            conn=self._create_tcp_connection(host,port)
            rpc=RPConn(0, open_socket=conn, **kwargs)
        else: raise RuisProtocolException("Illegal connspec: %s" % connspec)
        self._register_to_dir(str(rpc),rpc)
        rpc._cleanup_when_disconnected = [(self, str(rpc))]
        return str(rpc)

    def remove(self, conn_name):
        if not conn_name in self._contents:
            return None
        else:
            rpc = self._contents[conn_name]
            what_to_clean_up = []
            for (obj, name) in rpc._cleanup_when_disconnected:
                if name != conn_name and obj != self:
                    what_to_clean_up.append((obj, name))
            rpc._cleanup_when_disconnected = what_to_clean_up
            del self._contents[conn_name]
            if rpc._connection:
                rpc.close_connection()
            return conn_name

    def get_host(self, conn_name):
        return self._contents[conn_name].get_host()

    def connection(self, conn_name):
        return self._contents[conn_name]

    def _accept_tcp_connection(self,port,allow_host=''):
        """
        Returns an open socket object.
        
        Accepts one tcp connection to the given port. Special
        requirements for the connection are set with allow_host and
        client_key parameters. If allow_host is given, the connection
        must be established from the given host.
        """
        if port in self._sockets:
            s=self._sockets[port]
        else:
            s=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
            s.bind(('',port))
            s.listen(1)
            self._sockets[port]=s
        src=None
        # wait until (an allowed) connection is received
        while src==None or (allow_host!='' and src!=allow_host):
            conn,src=s.accept()
        return conn

    def _create_tcp_connection(self,host,port):
        # ??? TODO FIXME: what if something goes worng?
        c=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        #print "connecting to",host,port
        c.connect((host,int(port)))
        #print "connection established"
        return c

    def auto_accept(self,port,background=0):
        """
        Accept automatically every incoming connection.
        """
        def auto_accept_loop(port):
            while 1: self.accept_connection(port)

        if self._autoaccept==1: return
        self._autoaccept=1
        if background==0: auto_accept_loop(port)
        else: thread.start_new_thread(auto_accept_loop,(port,))

    def get_remote_caller(self):
        """
        Returns the remote caller, that is, the name of the /conn/rp
        entry from which the most recent remote call is originated, or
        None if there is no remote call in the stack.
        """
        # Look for RPConn._server code from stack
        stack = inspect.stack()
        rpconn_server_code = RPConn._server.func_code
        for stackitem in xrange(len(stack)-1, 0, -1):
            func_code = stack[stackitem][0].f_code
            if func_code == rpconn_server_code:
                # Check that the RPConn object is in this directory.
                # (If there are many ruis file systems within the same
                # process, the found connection might belong to some
                # other ruis protocol).
                conn_obj = stack[stackitem][0].f_locals['self']
                if self._contents.get(conn_obj.name, None) == conn_obj:
                    return conn_obj.name
        return None

_immutable_types = {bool:None, float:None, int:None, long:None, str:None, tuple:None, frozenset: None}

def _create_proxy_object(o):
    """
    This function returns an object that represents local object o in
    a remote peer. Immutables will be sent as they are (that is,
    passed by value) and proxies of mutables will be sent (that is,
    pass by reference). All returned values should be pickleable.
    """
    global _proxytable
    if isinstance(o, types.FunctionType) or isinstance(o, types.MethodType):
        if not id(o) in _proxytable: _proxytable[id(o)]=o
        return FuncProxy(id(o))
    elif type(o) in _immutable_types:
        return o
    elif hasattr(o,'__class__'):
        proxy_o = ProxyMsg(o)
        if hasattr(proxy_o,'remote_id'): # not a file-system object
            if not proxy_o.remote_id in _proxytable:
                _proxytable[proxy_o.remote_id] = o
        else:
            # Consider FIX: object 'a' is a file-system
            # object, which is located by remote_fullname
            # this is most likely stupid and should
            # perhaps be based on id too.
            pass
        return proxy_o
    else:
        raise Exception("WARNING: rp cannot handle argument of type %s." % (type(o),))
        # return o # pray that pickling works for o


def _create_methods(proxy,mcs):
    def create_call_forwarder(method_name,mcs):
        def call_forwarder(*args,**kwargs):
            global _proxytable
            # Convert parameters of the function call to proxies, if necessary
            # TODO: It probably requires writing proxies to some mutable
            # objects like list, dict, ...
            final_args=[_create_proxy_object(a) for a in args]
            # send the message
            # make synchronous call: wait for return value
            if hasattr(proxy, 'remote_id'):
                retval=mcs.request(ForwardedCallMsg(proxy.remote_id,
                                                    method_name,final_args,kwargs))
            else:
                retval=mcs.request(ForwardedCallMsg(proxy.remote_fullname[0],
                                                    method_name,final_args,kwargs))
            if type(retval)==ProxyMsg:
                _create_methods(retval,mcs)
            elif type(retval)==CarrierMsg:
                if type(retval.payload)==FuncProxy:
                    retval.payload.func_channel=mcs
                retval=retval.payload
            elif type(retval)==ConnectionLost:
                pass # ConnectionLost object can be a return value of a remote function
            else:
                raise Exception("Forwarded call returned: %s (value: '%s')" % (type(retval),retval))
            return retval
        return call_forwarder
    
    for m in proxy.method_names:
        setattr(proxy,m,create_call_forwarder(m,mcs))
        #print "Created proxy method",m


class RPConn(DirType):
    counter=0
    
    def __init__(self, am_i_server, name='', open_socket=None, local_root=None):
        """
        Implements /conn/rp/n
        
        This object is instantiated by /conn/rp.create()
        
        connspec is either an existing connection through which the
        rp connection is established, or string rp://host:port. If
        open_socket (an already open socket) is given, this class
        works on the server side.
        """
        # self._connection is request channel (my requests will be sent there)
        # self._server_connection is serving channel (read opposite requests)
        DirType.__init__(self)
        self._local_root=local_root
        if am_i_server:
            # work as a server
            self._connection=MultiChannelSocket(open_socket, 0, self._server)
        else:
            self._connection=MultiChannelSocket(open_socket, 1, self._server)
        if name=='': self.name=str(RPConn.counter) # FIX RACE on counter
        else: self.name=name
        RPConn.counter+=1

    def close_connection(self):
        # FIX: a nice shut down in multichannel would be nice...
        c=self._connection
        self._connection=None
        if c==None: return # another thread usually gives up here FIX RACE
        try: c._socket.shutdown(socket.SHUT_RDWR)
        except: pass
        c._socket.close()
        to_be_cleaned_up = self._cleanup_when_disconnected
        self._cleanup_when_disconnected = []
        for dirobj, name in to_be_cleaned_up:
            dirobj.remove(name)

    def get_host(self):
        return self._connection._socket.getpeername()[0]

    def get_port(self):
        return self._connection._socket.getpeername()[1]

    def __getitem__(self,dirspec):
        # When this is called, ruis file system wants to access the
        # directory. This access must be forwarded to the opposite
        # RPConn object.
        msg=self._connection.request(DirspecReqMsg(dirspec))
        if type(msg)==ProxyMsg:
            _create_methods(msg,self._connection) # create call forwarding methods
            return msg
        else:
            print "just received something strange. i hate this."

    def get_localitem(self, dirspec):
        #print "fetching",dirspec,"from local rootfs"
        return self._local_root[dirspec]

    def _server(self, mcs, msg, chn):
        """
        This method is the default_msg_handler of self._connection
        multichannelsocket (mcs). Incoming requests come this way.
        """
        if mcs==None:
            # the connection is lost!
            # it may be that self is already being removed (that is,
            # connection is most likely set to none). if it is not,
            # that is, there is a network error or something, remove
            # self now. CONSIDER/FIX RACE
            if self._connection!=None:
                self.close_connection()
        if type(msg)==DirspecReqMsg:
            # file system object requested. send a proxy
            # this proxy object is not stored in _proxytable because
            # it can be located by dirspec
            mcs.send(ProxyMsg(self.get_localitem(msg.dirspec)),chn)
        elif type(msg)==ForwardedCallMsg:
            # Cleaning up arguments:
            for a in msg.args:
                if type(a)==FuncProxy:
                    # There is a function proxy in the argument list,
                    # set the channel through which the proxy can
                    # communicate with the implementation of the
                    # function.
                    a.func_channel=mcs
                elif type(a)==ProxyMsg:
                    _create_methods(a,mcs)
            if msg.object_id!=None: # ForwardedCall to an object in ruis
                # debugmsg("ForwardedCall to an object in ruis: %s" % (msg.object_id,))
                if msg.object_id in _proxytable: # based on remote_id
                    obj=_proxytable[msg.object_id]
                else: # based on remote_fullname
                    obj=self._local_root[msg.object_id]
                method=getattr(obj,msg.method_name)
                if msg.method_name=='__getitem__':
                    # getitem called, sending proxy
                    mcs.send(ProxyMsg(method(*msg.args,**msg.kwargs)),chn)
                else:
                    # otherwise, sending pickled return value ...yet
                    # some types probably should be handled with more
                    # care in real life...

                    # return value of a forwarded call is carried by
                    # CarrierMsg
                    # Convert arguments...
                    
                    retval=method(*msg.args,**msg.kwargs)
                    if type(retval)==types.FunctionType:
                        if not id(retval) in _proxytable:
                            _proxytable[id(retval)]=retval
                        retval=FuncProxy(id(retval))
                    # add here: elif type(retval)==types.InstanceType...
                    # send a proxy and use _proxytable
                    mcs.send(CarrierMsg(retval),chn)
            else: # msg.object_id not set => ForwardedCall to a global function
                # print "calling local function",msg.method_name
                method=_proxytable[msg.method_name]
                # earlier implementation used global namespace, which
                # is a a bad idea (think of lambdas functions, for
                # example)
                retval=method(*msg.args,**msg.kwargs)
                # FIX: REPEATING CODE!!!
                # Implement and use _create_proxy_object(o)
                # WHY THIS BREAKS SHARE_TEST?: retval = _create_proxy_object(retval)
                if type(retval)==types.FunctionType:
                    if not id(retval) in _proxytable:
                        _proxytable[id(retval)]=retval
                    retval=FuncProxy(id(retval))
                mcs.send(CarrierMsg(retval),chn)
