# share.py - Ruis directory for location-independent object access
#
# Copyright 2008 Antti Kervinen (ask@cs.tut.fi)
#
# This file is part of Ruis.
#
# Ruis is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/lgpl.html>.

"""
This implements /share directory in the ruis file system. Contents of
the share directory are shown in /share directory in every peer that
is both connected using /conn/rp and has /share directory.
"""

from ruis import DirType, FileType, _split_dirspec, \
    ConnectionLost, debugmsg_printer
import thread

debugmsg = debugmsg_printer('SHARE')

class share(DirType):
    def __init__(self,local_root,*a,**kw):
        DirType.__init__(self,local_root,*a,**kw)
        self._root=local_root
        self._remote_shares={}
        self._rslock=thread.allocate_lock()
        local_root['/conn/rp/events/return:create'].set_callback(
            self._new_conn_callback)
        local_root['/conn/rp/events/return:remove'].set_callback(
            self._remove_conn_callback)

    def _new_conn_callback(self,return_value,*a,**kw):
        c=return_value
        if 'share' in self._root['/conn/rp/' + c].ls():
            self._rslock.acquire()
            debugmsg('Added to shares: %s' % (c,))
            self._remote_shares[c]=self._root['/conn/rp/'+c+'/share']
            self._rslock.release()

    def _remove_conn_callback(self, remove_retval, removed_name, **kw):
        if remove_retval == None: return
        if removed_name in self._remote_shares:
            self._rslock.acquire()
            debugmsg('removed from shares: %s' % (removed_name,))
            del self._remote_shares[removed_name]
            self._rslock.release()
        else:
            debugmsg("removed connection without shares: %s" % (removed_name,))

    def ls(self):
        files=self._contents.keys()
        self._rslock.acquire()
        shares=self._remote_shares.values()
        self._rslock.release()
        for rs in shares:
            remote_contents=rs.local_ls()
            if type(remote_contents)!=ConnectionLost:
                files.extend(remote_contents)
        files.sort()
        return tuple(files)

    def local_ls(self):
        return self._contents.keys()

    def __getitem__(self,dirspec):
        # Check if dirspec can be found locally
        fso=DirType.__getitem__(self,dirspec)
        if fso!=None: return fso

        # Next, try remote shares
        splitted=_split_dirspec(dirspec)
        self._rslock.acquire()
        conn_shares=[(c,rs) for c,rs in self._remote_shares.iteritems()]
        self._rslock.release()
        for c,rs in conn_shares:
            remote_contents=rs.local_ls()
            if type(remote_contents)==ConnectionLost: continue
            if splitted[0] in remote_contents:
                fso=self._root['/conn/rp/'+c+'/share/'+dirspec]
                if fso!=None: return fso
        return None # Could not find dirspec anywhere.
