#!/usr/bin/env python

# XSChat - Cross-site chat
#
# This example introduces web form protocol, /conn/webformp.
#
# Start a cross-site chat master with command
#
#     xschat HTTP-PORT
#
# and cross-site chat slaves with command
#
#     xschat HTTP-PORT tcp://MASTER-IP-ADDRESS:RP-PORT
#
# Slaves connect to the master with Ruis protocol. Both the master and
# the slaves run an http server at HTTP-PORT. Browsers connected to
# any of the http servers participate the same chat.
#
# Warning: some concurrency issues are happily omitted in this code
# for readability. For example, it can happen that two entries
# with the same name are created in /share/textareadir.

import ruis
import sys

fs = ruis.newfs('ruis_modules')

login_form_html = '''
<font size="5" color="blue">X</font>-site chat<br>
Login: <input name="username" type="text" />
<input name="ok-button" type="button" value="Ok" />'''

chatform_html = '''
<font size="5" color="blue">X</font>-site chat<br>
Logged in as <strong>%s</strong><br>
<textarea name="incoming" cols="60" rows="8"></textarea><br>
Write a message in the box and click it to send it:<br>
<input name="outgoing" type="text" size="60" /><br>
<input name="undo" type="button" value="Take back" />
'''

class textareadir(ruis.DirType):
    def __init__(self, *a, **kw):
        ruis.DirType.__init__(self, *a, **kw)
        self._buffer = [''] * 8
        
    def broadcast(self, msg):
        self._buffer = self._buffer[1:] + [msg]
        for ta in self._contents.values():
            ta.onchange('\n'.join(self._buffer))

    def takeback(self, prefix):
        self._buffer = [(msg, prefix + " [taken back]")[msg.startswith(prefix)]
                        for msg in self._buffer]
        self.broadcast("# %s took back his/her words" % prefix[:-1])

class session(object):
    def __init__(self, browserdir):
        self._browser = browserdir
        self._login_form = self._browser[ self._browser.create(login_form_html) ]
        self._login_form['ok-button/events/call:onclick'].set_callback(self.login)

    def login(self, *a, **kw):
        self._user = self._login_form['username'].get_value()
        self._chatform = self._browser[ self._browser.create(chatform_html % self._user) ]
        self._chatform['outgoing/events/call:onclick'].set_callback(self.send_msg)
        self._chatform['undo/events/call:onclick'].set_callback(self.take_it_back)
        fs['/share/textareadir'].create(self._chatform['incoming'],
                                  name=str(len(fs['/share/textareadir'].ls())))
        
        fs['/share/textareadir'].broadcast('# %s joins the chat' % self._user)
        
    def send_msg(self, msg, *a, **kw):
        if msg.strip() != "":
            fs['/share/textareadir'].broadcast('%s> %s' % (self._user, msg))
        self._chatform['outgoing'].onchange('')

    def take_it_back(self, *a, **kw):
        fs['/share/textareadir'].takeback('%s>' % self._user)

def main():
    RUISPORT = 1111
    if len(sys.argv)>=2:
        HTTPPORT = int(sys.argv[1])
    else:
        HTTPPORT = 7777

    if len(sys.argv)>=3: # run as slave
        print "xschat: rp connection to", sys.argv[2]
        fs['/conn/rp'].create(sys.argv[2])
    else:
        print "xschat: rp listens to", RUISPORT
        fs['/share'].create(textareadir())
        def accept_connections():
            while 1:
                fs['/conn/rp'].create('listen:tcp:1111')
        ruis.async(accept_connections)
    
    sessions = []

    while 1:
        print "xschat: webformp listens to", HTTPPORT
        browser_id = fs['/conn/webformp'].create('listen:tcp:%s' % HTTPPORT)
        if not browser_id: break

        print "xschat: session", browser_id, "established"
        new_session = session(fs['/conn/webformp/' + browser_id])
    
        sessions.append(new_session)

if __name__ == '__main__':
    main()
